# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Daemon subcommands: start, stop, logs, list, kill, add, remove."""

import collections
import itertools
import logging
import subprocess
import sys
import time
from typing import Optional

import typer

from elhaz.daemon import (
    Client,
    DaemonService,
    Server,
    configure_daemon_logging,
)
from elhaz.exceptions import ElhazDaemonError

from ..constants import state
from .output import print_error, print_success
from .prompts import resolve_name

app = typer.Typer(
    name="daemon",
    help="Manage the elhaz daemon.",
    no_args_is_help=True,
)


def _is_running() -> bool:
    """Return True if the daemon is reachable on the current socket."""

    try:
        with Client(state) as c:
            c.send("list")
        return True
    except ElhazDaemonError:
        return False


def _wait_until_running(timeout: float = 5.0) -> bool:
    """Poll until the daemon is reachable or *timeout* elapses.

    Parameters
    ----------
    timeout : float, optional
        Seconds to wait before giving up.

    Returns
    -------
    bool
        True if the daemon became reachable within *timeout*.
    """

    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        if _is_running():
            return True
        time.sleep(0.1)
    return False


def _wait_until_stopped(timeout: float = 5.0) -> bool:
    """Poll until the daemon is unreachable or *timeout* elapses.

    Parameters
    ----------
    timeout : float, optional
        Seconds to wait before giving up.

    Returns
    -------
    bool
        True if the daemon became unreachable within *timeout*.
    """

    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        if not _is_running():
            return True
        time.sleep(0.1)
    return False


def _daemon_subprocess_cmd() -> list[str]:
    """Build the ``elhaz daemon _serve`` subprocess command.

    Forwards all current Constants values so the daemon process uses
    the same configuration as the CLI invocation.

    Returns
    -------
    list[str]
        Argv for :func:`subprocess.Popen`.
    """

    return [
        sys.executable,
        "-m",
        "elhaz.cli",
        "--socket-path",
        str(state.socket_path),
        "--logging-path",
        str(state.daemon_logging_path),
        "--config-dir",
        str(state.config_dir),
        "--max-unix-socket-connections",
        str(state.max_unix_socket_connections),
        "--max-daemon-cache-size",
        str(state.max_daemon_cache_size),
        "daemon",
        "_serve",
    ]


@app.command("status")
def daemon_status() -> None:
    """Report whether the daemon is currently running.

    Exits with code 0 if the daemon is reachable, 1 if it is not.
    """

    if _is_running():
        typer.secho("Daemon is running.", fg=typer.colors.GREEN)
        typer.echo(f"  Socket: {state.socket_path}")
    else:
        typer.secho("Daemon is not running.", fg=typer.colors.YELLOW)
        raise typer.Exit(1)


@app.command("start")
def daemon_start() -> None:
    """Start the daemon in the background."""

    if _is_running():
        typer.echo("Daemon is already running.")
        return

    subprocess.Popen(
        _daemon_subprocess_cmd(),
        start_new_session=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        stdin=subprocess.DEVNULL,
    )

    if _wait_until_running():
        print_success("Daemon started.")
        typer.echo(f"  Socket:  {state.socket_path}")
        typer.echo(f"  Log:     {state.daemon_logging_path}")
    else:
        print_error(
            "Daemon did not start within 5 seconds. "
            f"Check the log file: {state.daemon_logging_path}"
        )
        raise typer.Exit(1)


@app.command("stop")
def daemon_stop() -> None:
    """Stop the running daemon gracefully."""

    if not _is_running():
        typer.echo("Daemon is not running.")
        return

    try:
        with Client(state) as client:
            response = client.send("kill")
    except ElhazDaemonError as exc:
        print_error(str(exc))
        raise typer.Exit(1)

    if not response.ok:
        print_error(
            response.error.message if response.error else "Unknown error."
        )
        raise typer.Exit(1)

    if _wait_until_stopped():
        print_success("Daemon stopped.")
    else:
        print_error("Daemon did not stop within 5 seconds.")
        raise typer.Exit(1)


@app.command("kill")
def daemon_kill() -> None:
    """Forcefully stop the daemon (alias for ``stop``)."""

    daemon_stop()


@app.command("logs")
def daemon_logs(
    tail: Optional[int] = typer.Option(
        50,
        "--tail",
        "-t",
        help="Show the last N lines (default: 50).",
    ),
    head: Optional[int] = typer.Option(
        None,
        "--head",
        "-h",
        help="Show the first N lines.",
    ),
) -> None:
    """Print daemon log output.

    Defaults to the last 50 lines. Use ``--head`` for the first N
    lines, or ``--tail 0`` to stream the entire file.
    """

    path = state.daemon_logging_path

    if not path.exists():
        typer.echo("No log file found.")
        return

    if head is not None:
        with open(path, encoding="utf-8") as f:
            for line in itertools.islice(f, head):
                typer.echo(line, nl=False)
    else:
        n = tail if tail and tail > 0 else None
        with open(path, encoding="utf-8") as f:
            for line in collections.deque(f, maxlen=n):
                typer.echo(line, nl=False)


@app.command("list")
def daemon_list() -> None:
    """List all active sessions in the daemon's cache."""

    try:
        with Client(state) as client:
            response = client.send("list")
    except ElhazDaemonError as exc:
        print_error(f"Daemon unreachable: {exc}")
        raise typer.Exit(1)

    if not response.ok:
        print_error(
            response.error.message if response.error else "Unknown error."
        )
        raise typer.Exit(1)

    sessions: list[str] = response.data or []
    if not sessions:
        typer.echo("No active sessions.")
        return
    for name in sessions:
        typer.echo(name)


@app.command("add")
def daemon_add(
    name: Optional[str] = typer.Option(
        None, "--name", "-n", help="Config name."
    ),
) -> None:
    """Initialize an AWS session and add it to the daemon's cache."""

    name = resolve_name(name, state, message="Select a config:")

    try:
        with Client(state) as client:
            response = client.send("add", {"config": name})
    except ElhazDaemonError as exc:
        print_error(f"Daemon unreachable: {exc}")
        raise typer.Exit(1)

    if not response.ok:
        print_error(
            response.error.message if response.error else "Unknown error."
        )
        raise typer.Exit(1)

    print_success(f"Session '{name}' added to daemon cache.")


@app.command("remove")
def daemon_remove(
    name: Optional[str] = typer.Option(
        None, "--name", "-n", help="Config name."
    ),
) -> None:
    """Remove an active session from the daemon's cache."""

    name = resolve_name(
        name,
        state,
        from_daemon=True,
        message="Select a session to remove:",
    )

    try:
        with Client(state) as client:
            response = client.send("remove", {"config": name})
    except ElhazDaemonError as exc:
        print_error(f"Daemon unreachable: {exc}")
        raise typer.Exit(1)

    if not response.ok:
        print_error(
            response.error.message if response.error else "Unknown error."
        )
        raise typer.Exit(1)

    print_success(f"Session '{name}' removed from daemon cache.")


@app.command("_serve", hidden=True)
def daemon_serve() -> None:
    """Start the daemon server (internal; invoked by ``daemon start``)."""

    configure_daemon_logging(state)

    _log = logging.getLogger(__name__)

    try:
        service = DaemonService(max_size=state.max_daemon_cache_size)
        server = Server(state, service)
        server.run()
    except Exception:
        # stderr is /dev/null in the detached subprocess, so any startup
        # or runtime exception must be written to the log file explicitly.
        _log.exception("Daemon process failed.")
        raise



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Config subcommands: add, list, get, update, remove."""

import os
import subprocess
from dataclasses import dataclass
from dataclasses import field as dc_field
from typing import Literal, Optional

import typer

from elhaz.config import Config
from elhaz.exceptions import (
    ElhazAlreadyExistsError,
    ElhazNotFoundError,
    ElhazValidationError,
)

from ..constants import state
from .output import obscure, print_error, print_json, print_success
from .prompts import (
    ask_text,
    ask_yes_no,
    list_local_configs,
    resolve_name,
    select_local_config,
)

app = typer.Typer(
    name="config",
    help="Manage elhaz configurations.",
    no_args_is_help=True,
)

FieldKind = Literal["scalar", "str_list", "model_list"]


@dataclass
class FieldDescriptor:
    """Descriptor for a single interactive config field.

    Parameters
    ----------
    key : str
        Dict key / Pydantic field name.
    label : str
        Human-readable prompt label.
    kind : FieldKind
        How to collect the value: ``scalar``, ``str_list``, or
        ``model_list``.
    cast : type
        Cast function applied to scalar input. Ignored for list kinds.
    hint : str or None
        Inline hint appended to the scalar prompt (e.g. ``"seconds"``).
    sub_fields : list[FieldDescriptor]
        Sub-field descriptors for ``model_list`` kinds.
    """

    key: str
    label: str
    kind: FieldKind = "scalar"
    cast: type = str
    hint: str | None = None
    sub_fields: list["FieldDescriptor"] = dc_field(default_factory=list)


_ASSUME_ROLE_OPTIONAL: list[FieldDescriptor] = [
    FieldDescriptor("RoleSessionName", "RoleSessionName"),
    FieldDescriptor(
        "DurationSeconds", "DurationSeconds", cast=int, hint="seconds"
    ),
    FieldDescriptor("ExternalId", "ExternalId"),
    FieldDescriptor("SerialNumber", "SerialNumber (MFA device ARN)"),
    FieldDescriptor("TokenCode", "TokenCode (MFA token)"),
    FieldDescriptor("SourceIdentity", "SourceIdentity"),
    FieldDescriptor("Policy", "Inline session policy (JSON string)"),
    FieldDescriptor(
        "PolicyArns", "PolicyArns", kind="str_list", hint="policy ARN"
    ),
    FieldDescriptor(
        "Tags",
        "Tags",
        kind="model_list",
        sub_fields=[
            FieldDescriptor("Key", "Key"),
            FieldDescriptor("Value", "Value"),
        ],
    ),
    FieldDescriptor(
        "TransitiveTagKeys",
        "TransitiveTagKeys",
        kind="str_list",
        hint="tag key",
    ),
    FieldDescriptor(
        "ProvidedContexts",
        "ProvidedContexts",
        kind="model_list",
        sub_fields=[
            FieldDescriptor("ProviderArn", "ProviderArn"),
            FieldDescriptor("ContextAssertion", "ContextAssertion"),
        ],
    ),
]

_STS_FIELDS: list[FieldDescriptor] = [
    FieldDescriptor("region_name", "Region name"),
    FieldDescriptor("api_version", "API version"),
    FieldDescriptor("use_ssl", "Use SSL?", cast=bool),
    FieldDescriptor("verify", "Verify SSL?", cast=bool),
    FieldDescriptor("endpoint_url", "STS endpoint URL"),
    FieldDescriptor("aws_access_key_id", "AWS access key ID"),
    FieldDescriptor("aws_secret_access_key", "AWS secret access key"),
    FieldDescriptor("aws_session_token", "AWS session token"),
    FieldDescriptor("aws_account_id", "AWS account ID"),
]

_SESSION_FIELDS: list[FieldDescriptor] = [
    FieldDescriptor("region_name", "Region name"),
    FieldDescriptor("profile_name", "AWS profile name"),
    FieldDescriptor("aws_account_id", "AWS account ID"),
    FieldDescriptor("aws_access_key_id", "AWS access key ID"),
    FieldDescriptor("aws_secret_access_key", "AWS secret access key"),
    FieldDescriptor("aws_session_token", "AWS session token"),
]


def _collect_str_list(
    descriptor: FieldDescriptor,
    existing: list[str] | None = None,
) -> list[str] | None:
    """Interactively collect a list of strings for *descriptor*.

    Parameters
    ----------
    descriptor : FieldDescriptor
        Field being collected.
    existing : list[str] or None, optional
        Current values shown when updating a config.

    Returns
    -------
    list[str] or None
        Collected items, or None if the user entered nothing.
    """

    if existing:
        typer.echo(f"  Current {descriptor.label}: {existing}")
        if not ask_yes_no(
            f"  Replace existing {descriptor.label}?", default=False
        ):
            return existing

    hint = f" ({descriptor.hint})" if descriptor.hint else ""
    typer.echo(f"  Enter {descriptor.label}{hint} — empty line to finish:")
    items: list[str] = []
    i = 1
    while True:
        val = ask_text(f"  [{i}]:", default="")
        if not val:
            break
        items.append(val)
        i += 1
    return items or None


def _collect_model_list(
    descriptor: FieldDescriptor,
    existing: list[dict] | None = None,
) -> list[dict] | None:
    """Interactively collect a list of sub-field dicts for *descriptor*.

    Parameters
    ----------
    descriptor : FieldDescriptor
        Field being collected. ``sub_fields`` must be non-empty.
    existing : list[dict] or None, optional
        Current values shown when updating a config.

    Returns
    -------
    list[dict] or None
        Collected items, or None if the user skipped.
    """

    if existing:
        typer.echo(f"  Current {descriptor.label}:")
        for entry in existing:
            typer.echo(f"    {entry}")
        if not ask_yes_no(
            f"  Replace existing {descriptor.label}?", default=False
        ):
            return existing

    items: list[dict] = []
    while True:
        entry: dict = {}
        typer.echo(f"  New {descriptor.label} item:")
        for sub in descriptor.sub_fields:
            val = ask_text(f"    {sub.label} (required):", required=True)
            entry[sub.key] = val
        items.append(entry)
        if not ask_yes_no(
            f"  Add another {descriptor.label} item?", default=False
        ):
            break
    return items or None


def _collect_optional(
    fields: list[FieldDescriptor],
    existing: dict | None = None,
) -> dict:
    """Prompt for a list of optional fields, skipping blanks.

    Parameters
    ----------
    fields : list[FieldDescriptor]
        Field descriptors.
    existing : dict or None, optional
        Pre-populate prompts with existing values.

    Returns
    -------
    dict
        Non-empty field values keyed by field name.
    """

    result: dict = {}
    existing = existing or {}
    for desc in fields:
        if desc.kind == "scalar":
            default = str(existing.get(desc.key, ""))
            skip = " (optional — Enter to skip)"
            suffix = f" ({desc.hint})" if desc.hint else skip
            text = ask_text(f"{desc.label}{suffix}:", default=default)
            if text:
                try:
                    result[desc.key] = desc.cast(text)
                except (ValueError, TypeError):
                    typer.secho(
                        f"  Invalid value for {desc.key!r}; skipping.",
                        fg=typer.colors.YELLOW,
                        err=True,
                    )
        elif desc.kind == "str_list":
            if ask_yes_no(f"Configure {desc.label}?", default=False):
                collected = _collect_str_list(desc, existing.get(desc.key))
                if collected:
                    result[desc.key] = collected
        elif desc.kind == "model_list":
            if ask_yes_no(f"Configure {desc.label}?", default=False):
                collected = _collect_model_list(desc, existing.get(desc.key))
                if collected:
                    result[desc.key] = collected
    return result


def _build_config_interactively(
    existing: dict | None = None,
) -> dict:
    """Walk the user through constructing a full config payload.

    Parameters
    ----------
    existing : dict or None, optional
        Existing config values for pre-population (used by ``update``).

    Returns
    -------
    dict
        Config payload ready to pass to :meth:`~elhaz.config.Config.add`.
    """

    existing = existing or {}
    existing_ar = existing.get("AssumeRole", {})

    role_arn = ask_text(
        "RoleArn (required):",
        default=existing_ar.get("RoleArn", ""),
        required=True,
    )
    assume_role: dict = {"RoleArn": role_arn}

    if ask_yes_no("Configure optional AssumeRole settings?", default=False):
        assume_role.update(
            _collect_optional(_ASSUME_ROLE_OPTIONAL, existing_ar)
        )

    config: dict = {"AssumeRole": assume_role}

    if ask_yes_no("Configure STS client settings?", default=False):
        sts = _collect_optional(_STS_FIELDS, existing.get("STS", {}))
        if sts:
            config["STS"] = sts

    if ask_yes_no("Configure MFA settings?", default=False):
        cmd = ask_text(
            "MFA command (shell command or executable):",
            required=True,
        )
        timeout_str = ask_text("MFA timeout in seconds:", default="30")
        try:
            timeout = int(timeout_str) if timeout_str else 30
        except ValueError:
            timeout = 30
        config["MFA"] = {"command": cmd, "timeout": timeout}

    if ask_yes_no("Configure Session settings?", default=False):
        session = _collect_optional(
            _SESSION_FIELDS, existing.get("Session", {})
        )
        if session:
            config["Session"] = session

    return config


def _open_in_editor(path: "os.PathLike[str]") -> None:
    """Open *path* in the user's ``$EDITOR``, defaulting to ``vi``.

    Parameters
    ----------
    path : os.PathLike[str]
        File to open.
    """

    editor = os.environ.get("EDITOR", "vi")
    subprocess.run([editor, str(path)], check=False)


@app.command("add")
def config_add(
    name: Optional[str] = typer.Option(
        None, "--name", "-n", help="Config name."
    ),
) -> None:
    """Create a new config in the local config store."""

    if not name:
        name = ask_text("Config name:", required=True)

    cfg = Config(name, state)

    if cfg.file_path.exists():
        if not ask_yes_no(
            f"Config '{name}' already exists. Overwrite?",
            default=False,
        ):
            raise typer.Exit(0)
        cfg.delete()

    if ask_yes_no("Create config interactively?"):
        payload = _build_config_interactively()
        try:
            cfg.add(payload)
            print_success(f"Config '{name}' created.")
        except ElhazAlreadyExistsError as exc:
            print_error(str(exc))
            raise typer.Exit(1)
        except ElhazValidationError as exc:
            print_error(str(exc))
            raise typer.Exit(1)
    else:
        cfg.file_path.parent.mkdir(parents=True, exist_ok=True)
        cfg.file_path.touch()
        _open_in_editor(cfg.file_path)
        typer.echo(f"Config file saved to {cfg.file_path}.")


@app.command("list")
def config_list() -> None:
    """List all config names in the local config store."""

    names = list_local_configs(state)
    if not names:
        typer.echo("No configs found.")
        return
    for name in names:
        typer.echo(name)


@app.command("get")
def config_get(
    name: Optional[str] = typer.Option(
        None, "--name", "-n", help="Config name."
    ),
    obscure_values: bool = typer.Option(
        False,
        "--obscure",
        "-o",
        help="Redact sensitive field values (RoleArn, credentials, etc.).",
    ),
) -> None:
    """Return config details as formatted JSON."""

    if not name:
        names = list_local_configs(state)
        if not names:
            typer.secho(
                "No configs found. Run 'elhaz config add' first.",
                fg=typer.colors.YELLOW,
                err=True,
            )
            raise typer.Exit(1)
        name = select_local_config(state, "Select a config:")

    cfg = Config(name, state)

    if not cfg.file_path.exists():
        if ask_yes_no(
            f"Config '{name}' does not exist in the local config "
            "store. Would you like to create it?",
            default=False,
        ):
            config_add(name=name)
        raise typer.Exit(0)

    try:
        data = cfg.get()
        if obscure_values:
            data = obscure(data)
        print_json(data)
    except ElhazNotFoundError as exc:
        print_error(str(exc))
        raise typer.Exit(1)
    except ElhazValidationError as exc:
        print_error(str(exc))
        raise typer.Exit(1)


@app.command("update")
def config_update(
    name: Optional[str] = typer.Option(
        None, "--name", "-n", help="Config name."
    ),
) -> None:
    """Update a config interactively."""

    name = resolve_name(name, state, message="Select a config to update:")
    cfg = Config(name, state)

    try:
        existing = cfg.get()
    except ElhazNotFoundError:
        print_error(f"Config '{name}' not found.")
        raise typer.Exit(1)
    except ElhazValidationError as exc:
        print_error(str(exc))
        raise typer.Exit(1)

    if ask_yes_no("Edit interactively?"):
        payload = _build_config_interactively(existing)
        try:
            cfg.delete()
            cfg.add(payload)
            print_success(f"Config '{name}' updated.")
        except ElhazValidationError as exc:
            print_error(str(exc))
            raise typer.Exit(1)
    else:
        _open_in_editor(cfg.file_path)
        typer.echo(f"Config file saved to {cfg.file_path}.")


@app.command("remove")
def config_remove(
    name: Optional[str] = typer.Option(
        None, "--name", "-n", help="Config name."
    ),
) -> None:
    """Remove a config from the local config store."""

    name = resolve_name(name, state, message="Select a config to remove:")
    cfg = Config(name, state)

    if not cfg.file_path.exists():
        print_error(f"Config '{name}' not found.")
        raise typer.Exit(1)

    if not ask_yes_no(
        f"Remove config '{name}'? This cannot be undone.",
        default=False,
    ):
        raise typer.Exit(0)

    cfg.delete()
    print_success(f"Config '{name}' removed.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Pydantic models for boto3-refresh-session configuration."""

__all__ = [
    "AssumeRoleModel",
    "ConfigModel",
    "CredentialProcessModel",
    "ErrorModel",
    "MFAModel",
    "RequestModel",
    "ResponseModel",
    "SessionModel",
    "STSModel",
]

from typing import Any, Dict, List, Literal, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict

actions = Literal["add", "credentials", "kill", "list", "remove", "whoami"]


class _BaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class TagModel(_BaseModel):
    """Model for AWS STS session tags."""

    Key: str
    Value: str


class ProvidedContextModel(_BaseModel):
    """Model for AWS STS provided contexts."""

    ProviderArn: str
    ContextAssertion: str


class AssumeRoleModel(_BaseModel):
    """Model for AWS STS AssumeRole parameters."""

    RoleArn: str
    RoleSessionName: Optional[str] = None
    PolicyArns: Optional[List[str]] = None
    Policy: Optional[str] = None
    DurationSeconds: Optional[int] = None
    ExternalId: Optional[str] = None
    SerialNumber: Optional[str] = None
    TokenCode: Optional[str] = None
    Tags: Optional[List[TagModel]] = None
    TransitiveTagKeys: Optional[List[str]] = None
    SourceIdentity: Optional[str] = None
    ProvidedContexts: Optional[List[ProvidedContextModel]] = None


class STSModel(_BaseModel):
    """Model for AWS STS client parameters."""

    region_name: Optional[str] = None
    api_version: Optional[str] = None
    use_ssl: Optional[bool] = None
    verify: Optional[bool | str] = None
    endpoint_url: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_session_token: Optional[str] = None
    aws_account_id: Optional[str] = None


class MFAModel(_BaseModel):
    """Model for AWS STS MFA parameters."""

    command: str | List[str]
    timeout: Optional[int] = 30


class SessionModel(_BaseModel):
    """Model for AWS STS session parameters."""

    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_session_token: Optional[str] = None
    region_name: Optional[str] = None
    profile_name: Optional[str] = None
    aws_account_id: Optional[str] = None


class ConfigModel(_BaseModel):
    """Model for boto3-refresh-session configuration."""

    AssumeRole: AssumeRoleModel
    STS: Optional[STSModel] = None
    MFA: Optional[MFAModel] = None
    Session: Optional[SessionModel] = None


class CredentialProcessModel(BaseModel):
    """JSON shape required by AWS ``credential_process``.

    Attributes
    ----------
    Version : int
        Credential process protocol version. Always 1.
    AccessKeyId : str
        AWS access key ID.
    SecretAccessKey : str
        AWS secret access key.
    SessionToken : str
        AWS session token.
    Expiration : str
        ISO 8601 expiration timestamp.
    """

    Version: int = 1
    AccessKeyId: str
    SecretAccessKey: str
    SessionToken: str
    Expiration: str


class ErrorModel(_BaseModel):
    """Model for structured daemon error responses.

    Attributes
    ----------
    code : int
        Numeric error code (HTTP-like: 400, 404, 409, 500).
    message : str
        Human-readable description of the error.
    """

    code: int
    message: str


class RequestModel(_BaseModel):
    """Model for requests sent to the daemon.

    Attributes
    ----------
    version : int
        Protocol version. Defaults to 1.
    request_id : UUID
        Client-generated unique identifier echoed back in the response.
    action : str
        The daemon action to invoke.
    payload : dict[str, Any]
        Action-specific fields. Defaults to an empty dict.
    """

    version: int = 1
    request_id: UUID
    action: actions
    payload: Dict[str, Any] = {}


class ResponseModel(_BaseModel):
    """Model for responses sent by the daemon.

    Either ``data`` or ``error`` is populated, never both.

    Attributes
    ----------
    version : int
        Protocol version. Defaults to 1.
    request_id : UUID
        Echoed from the originating request.
    ok : bool
        True on success, False on error.
    data : Any, optional
        Action-specific result payload. Present when ``ok`` is True.
    error : ErrorModel, optional
        Structured error detail. Present when ``ok`` is False.
    """

    version: int = 1
    request_id: UUID
    ok: bool
    data: Any = None
    error: Optional[ErrorModel] = None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for elhaz.daemon helpers and DaemonService."""

from __future__ import annotations

import logging
import threading
from logging.handlers import RotatingFileHandler
from pathlib import Path
from uuid import uuid4

import pytest

from elhaz.constants import Constants
from elhaz.daemon import DaemonService, _error_code, configure_daemon_logging
from elhaz.exceptions import (
    BaseElhazError,
    ElhazAlreadyExistsError,
    ElhazBadRequestError,
    ElhazNotFoundError,
)
from elhaz.models import RequestModel


@pytest.fixture(autouse=True)
def _patch_session(monkeypatch):
    """Replace elhaz.daemon.Session with FakeSession."""
    from tests.conftest import FakeSession

    monkeypatch.setattr("elhaz.daemon.Session", FakeSession)


@pytest.fixture()
def service() -> DaemonService:
    return DaemonService()


def test_error_code_bad_request() -> None:
    exc = ElhazBadRequestError("bad")
    assert _error_code(exc) == 400


def test_error_code_not_found() -> None:
    exc = ElhazNotFoundError("nope")
    assert _error_code(exc) == 404


def test_error_code_already_exists() -> None:
    exc = ElhazAlreadyExistsError("dup")
    assert _error_code(exc) == 409


def test_error_code_base_returns_500() -> None:
    exc = BaseElhazError("generic")
    assert _error_code(exc) == 500


def test_error_code_unknown_subclass_returns_500() -> None:
    class _NewError(BaseElhazError):
        pass

    exc = _NewError("novel")
    assert _error_code(exc) == 500


def test_configure_daemon_logging_creates_parent_dir(
    tmp_path: Path,
    clean_elhaz_logger,
) -> None:
    c = Constants()
    c.daemon_logging_path = tmp_path / "logs" / "sub" / "daemon.log"
    assert not c.daemon_logging_path.parent.exists()
    configure_daemon_logging(c)
    assert c.daemon_logging_path.parent.exists()


def test_configure_daemon_logging_adds_one_rotating_handler(
    tmp_path: Path,
    clean_elhaz_logger,
) -> None:
    c = Constants()
    c.daemon_logging_path = tmp_path / "logs" / "daemon.log"
    configure_daemon_logging(c)
    rfhs = [
        h
        for h in clean_elhaz_logger.handlers
        if isinstance(h, RotatingFileHandler)
    ]
    assert len(rfhs) == 1


def test_configure_daemon_logging_idempotent(
    tmp_path: Path,
    clean_elhaz_logger,
) -> None:
    c = Constants()
    c.daemon_logging_path = tmp_path / "logs" / "daemon.log"
    configure_daemon_logging(c)
    configure_daemon_logging(c)  # second call
    rfhs = [
        h
        for h in clean_elhaz_logger.handlers
        if isinstance(h, RotatingFileHandler)
    ]
    assert len(rfhs) == 1


def test_configure_daemon_logging_preserves_stream_handler(
    tmp_path: Path,
    clean_elhaz_logger,
) -> None:
    stream_handler = logging.StreamHandler()
    clean_elhaz_logger.addHandler(stream_handler)

    c = Constants()
    c.daemon_logging_path = tmp_path / "logs" / "daemon.log"
    configure_daemon_logging(c)

    assert stream_handler in clean_elhaz_logger.handlers


def test_configure_daemon_logging_handler_path(
    tmp_path: Path,
    clean_elhaz_logger,
) -> None:
    c = Constants()
    log_path = tmp_path / "daemon.log"
    c.daemon_logging_path = log_path
    configure_daemon_logging(c)
    rfhs = [
        h
        for h in clean_elhaz_logger.handlers
        if isinstance(h, RotatingFileHandler)
    ]
    assert len(rfhs) == 1
    assert Path(rfhs[0].baseFilename) == log_path


def test_configure_daemon_logging_sets_level_info(
    tmp_path: Path,
    clean_elhaz_logger,
) -> None:
    c = Constants()
    c.daemon_logging_path = tmp_path / "daemon.log"
    configure_daemon_logging(c)
    assert clean_elhaz_logger.level == logging.INFO


def test_daemon_service_add_returns_config_name(
    service: DaemonService,
) -> None:
    result = service.add("demo")
    assert result == {"config": "demo"}


def test_daemon_service_add_puts_session_in_cache(
    service: DaemonService,
) -> None:
    service.add("demo")
    sessions = service.list()
    assert "demo" in sessions


def test_daemon_service_add_same_name_twice_overwrites(
    service: DaemonService,
) -> None:
    service.add("demo")
    service.add("demo")
    assert service.list().count("demo") == 1


def test_daemon_service_credentials_returns_dict(
    service: DaemonService,
) -> None:
    service.add("demo")
    creds = service.credentials("demo")
    assert isinstance(creds, dict)
    assert "access_key" in creds


def test_daemon_service_credentials_unknown_raises_not_found(
    service: DaemonService,
) -> None:
    with pytest.raises(ElhazNotFoundError):
        service.credentials("ghost")


def test_daemon_service_list_empty(service: DaemonService) -> None:
    assert service.list() == []


def test_daemon_service_list_returns_names(service: DaemonService) -> None:
    service.add("alpha")
    service.add("beta")
    result = service.list()
    assert "alpha" in result
    assert "beta" in result


def test_daemon_service_list_order(service: DaemonService) -> None:
    service.add("first")
    service.add("second")
    result = service.list()
    assert result.index("first") < result.index("second")


def test_daemon_service_remove_returns_config_name(
    service: DaemonService,
) -> None:
    service.add("demo")
    result = service.remove("demo")
    assert result == {"config": "demo"}


def test_daemon_service_remove_session_gone_from_cache(
    service: DaemonService,
) -> None:
    service.add("demo")
    service.remove("demo")
    assert "demo" not in service.list()


def test_daemon_service_remove_unknown_raises_not_found(
    service: DaemonService,
) -> None:
    with pytest.raises(ElhazNotFoundError):
        service.remove("ghost")


def test_daemon_service_whoami_returns_identity_without_response_metadata(
    service: DaemonService,
) -> None:
    service.add("demo")
    result = service.whoami("demo")
    assert "ResponseMetadata" not in result
    assert "Account" in result


def test_daemon_service_whoami_unknown_raises_not_found(
    service: DaemonService,
) -> None:
    with pytest.raises(ElhazNotFoundError):
        service.whoami("ghost")


def _make_request(action: str, payload: dict | None = None) -> RequestModel:
    return RequestModel(
        request_id=uuid4(),
        action=action,  # type: ignore[arg-type]
        payload=payload or {},
    )


def test_dispatch_routes_add(service: DaemonService) -> None:
    req = _make_request("add", {"config": "demo"})
    result = service.dispatch(req)
    assert result == {"config": "demo"}


def test_dispatch_routes_list(service: DaemonService) -> None:
    req = _make_request("list")
    result = service.dispatch(req)
    assert isinstance(result, list)


def test_dispatch_routes_credentials(service: DaemonService) -> None:
    service.add("demo")
    req = _make_request("credentials", {"config": "demo"})
    result = service.dispatch(req)
    assert "access_key" in result


def test_dispatch_routes_remove(service: DaemonService) -> None:
    service.add("demo")
    req = _make_request("remove", {"config": "demo"})
    result = service.dispatch(req)
    assert result == {"config": "demo"}


def test_dispatch_routes_whoami(service: DaemonService) -> None:
    service.add("demo")
    req = _make_request("whoami", {"config": "demo"})
    result = service.dispatch(req)
    assert "Account" in result


def test_dispatch_raises_bad_request_for_unknown_action(
    service: DaemonService,
) -> None:
    # Build a request object directly since the Literal type won't accept
    # "bogus" — use model_construct to bypass Pydantic validation.
    req = RequestModel.model_construct(
        request_id=uuid4(), action="bogus", payload={}, version=1
    )
    with pytest.raises(ElhazBadRequestError):
        service.dispatch(req)


def test_dispatch_raises_bad_request_for_kill_action(
    service: DaemonService,
) -> None:
    req = RequestModel.model_construct(
        request_id=uuid4(), action="kill", payload={}, version=1
    )
    with pytest.raises(ElhazBadRequestError):
        service.dispatch(req)


def test_dispatch_raises_bad_request_when_config_missing_for_add(
    service: DaemonService,
) -> None:
    req = _make_request("add", {})
    with pytest.raises(ElhazBadRequestError):
        service.dispatch(req)


def test_dispatch_raises_bad_request_when_config_missing_for_credentials(
    service: DaemonService,
) -> None:
    req = _make_request("credentials", {})
    with pytest.raises(ElhazBadRequestError):
        service.dispatch(req)


def test_daemon_service_concurrent_add_and_list(
    service: DaemonService,
) -> None:
    errors: list[Exception] = []

    def _adder():
        try:
            for i in range(10):
                service.add(f"session-{i}")
        except Exception as exc:
            errors.append(exc)

    def _lister():
        try:
            for _ in range(10):
                _ = service.list()
        except Exception as exc:
            errors.append(exc)

    t1 = threading.Thread(target=_adder)
    t2 = threading.Thread(target=_lister)
    t1.start()
    t2.start()
    t1.join()
    t2.join()

    assert errors == [], f"Errors during concurrent access: {errors}"
    # all sessions were added (cache may be bounded, but no exceptions)
    for i in range(min(10, service._cache.max_size)):
        pass  # just verify no exceptions occurred



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for elhaz.constants.Constants."""

from __future__ import annotations

from pathlib import Path

import pytest

from elhaz.constants import Constants
from elhaz.exceptions import ElhazValidationError


def test_config_dir_accepts_valid_path(tmp_path: Path) -> None:
    c = Constants()
    c.config_dir = tmp_path / "configs"
    assert c.config_dir == tmp_path / "configs"


def test_config_dir_rejects_none() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.config_dir = None  # type: ignore[assignment]


def test_config_dir_rejects_empty_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.config_dir = ""  # type: ignore[assignment]


def test_config_dir_rejects_string_path() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.config_dir = "/tmp/configs"  # type: ignore[assignment]


def test_config_dir_round_trip(tmp_path: Path) -> None:
    c = Constants()
    new_path = tmp_path / "myconfigs"
    c.config_dir = new_path
    assert c.config_dir == new_path


def test_socket_path_accepts_valid_path(tmp_path: Path) -> None:
    c = Constants()
    c.socket_path = tmp_path / "daemon.sock"
    assert c.socket_path == tmp_path / "daemon.sock"


def test_socket_path_rejects_none() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.socket_path = None  # type: ignore[assignment]


def test_socket_path_rejects_empty_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.socket_path = ""  # type: ignore[assignment]


def test_socket_path_rejects_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.socket_path = "/tmp/daemon.sock"  # type: ignore[assignment]


def test_socket_path_round_trip(tmp_path: Path) -> None:
    c = Constants()
    p = tmp_path / "sock" / "daemon.sock"
    c.socket_path = p
    assert c.socket_path == p


def test_daemon_logging_path_accepts_valid_path(tmp_path: Path) -> None:
    c = Constants()
    c.daemon_logging_path = tmp_path / "daemon.log"
    assert c.daemon_logging_path == tmp_path / "daemon.log"


def test_daemon_logging_path_rejects_none() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.daemon_logging_path = None  # type: ignore[assignment]


def test_daemon_logging_path_rejects_empty_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.daemon_logging_path = ""  # type: ignore[assignment]


def test_daemon_logging_path_rejects_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.daemon_logging_path = "/tmp/daemon.log"  # type: ignore[assignment]


def test_daemon_logging_path_round_trip(tmp_path: Path) -> None:
    c = Constants()
    p = tmp_path / "logs" / "daemon.log"
    c.daemon_logging_path = p
    assert c.daemon_logging_path == p


def test_max_unix_socket_connections_accepts_one() -> None:
    c = Constants()
    c.max_unix_socket_connections = 1
    assert c.max_unix_socket_connections == 1


def test_max_unix_socket_connections_accepts_large_int() -> None:
    c = Constants()
    c.max_unix_socket_connections = 100
    assert c.max_unix_socket_connections == 100


def test_max_unix_socket_connections_rejects_zero() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.max_unix_socket_connections = 0


def test_max_unix_socket_connections_rejects_negative() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.max_unix_socket_connections = -1


def test_max_unix_socket_connections_rejects_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.max_unix_socket_connections = "5"  # type: ignore[assignment]


def test_max_unix_socket_connections_round_trip() -> None:
    c = Constants()
    c.max_unix_socket_connections = 7
    assert c.max_unix_socket_connections == 7


def test_max_daemon_cache_size_accepts_one() -> None:
    c = Constants()
    c.max_daemon_cache_size = 1
    assert c.max_daemon_cache_size == 1


def test_max_daemon_cache_size_accepts_ten() -> None:
    c = Constants()
    c.max_daemon_cache_size = 10
    assert c.max_daemon_cache_size == 10


def test_max_daemon_cache_size_rejects_zero() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.max_daemon_cache_size = 0


def test_max_daemon_cache_size_rejects_negative() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.max_daemon_cache_size = -1


def test_max_daemon_cache_size_rejects_string() -> None:
    c = Constants()
    with pytest.raises(ElhazValidationError):
        c.max_daemon_cache_size = "10"  # type: ignore[assignment]


def test_max_daemon_cache_size_round_trip() -> None:
    c = Constants()
    c.max_daemon_cache_size = 3
    assert c.max_daemon_cache_size == 3


def test_config_file_extension_is_yaml() -> None:
    c = Constants()
    assert c.config_file_extension == ".yaml"


def test_config_file_extension_has_no_setter() -> None:
    c = Constants()
    with pytest.raises(AttributeError):
        c.config_file_extension = ".json"  # type: ignore[misc]


def test_multiple_instances_are_independent(tmp_path: Path) -> None:
    a = Constants()
    b = Constants()
    a.config_dir = tmp_path / "a_configs"
    assert b.config_dir != a.config_dir


def test_instance_defaults_unchanged_after_sibling_mutation(
    tmp_path: Path,
) -> None:
    a = Constants()
    b = Constants()
    default_dir = b.config_dir
    a.config_dir = tmp_path / "x"
    assert b.config_dir == default_dir



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["Constants", "state"]

from pathlib import Path

from elhaz.exceptions import ElhazValidationError


class Constants:
    """Constants for config file discovery and management and daemon
    configuration.

    Attributes
    ----------
    config_dir : Path
        The directory where config files are stored. Default is
        ~/.elhaz/configs.
    config_file_extension : str
        The file extension for config files. Default is .yaml.
    socket_path : Path
        The path to the Unix socket for daemon communication. Default is
        ~/.elhaz/sock/daemon.sock.
    daemon_logging_path : Path
        The path to the log file for the daemon. Default is
        ~/.elhaz/logs/daemon.log.
    max_unix_socket_connections : int
        The maximum number of pending Unix socket connections the daemon will
        allow in its listen backlog. Default is 5.
    max_daemon_cache_size : int
        The maximum number of sessions the daemon session cache will retain
        before evicting the least recently used entry. Default is 10.
    """

    _config_dir: Path = Path.home() / ".elhaz/configs"
    _config_file_extension: str = ".yaml"
    _socket_path: Path = Path.home() / ".elhaz" / "sock" / "daemon.sock"
    _daemon_logging_path: Path = Path.home() / ".elhaz/logs/daemon.log"
    _max_unix_socket_connections: int = 5
    _max_daemon_cache_size: int = 10

    @property
    def config_dir(self) -> Path:
        return self._config_dir

    @config_dir.setter
    def config_dir(self, value: Path) -> None:
        if not value or not isinstance(value, Path):
            raise ElhazValidationError(f"Invalid config directory: '{value}'")
        self._config_dir = value

    @property
    def config_file_extension(self) -> str:
        return self._config_file_extension

    @property
    def socket_path(self) -> Path:
        return self._socket_path

    @socket_path.setter
    def socket_path(self, value: Path) -> None:
        if not value or not isinstance(value, Path):
            raise ElhazValidationError(f"Invalid socket path: '{value}'")
        self._socket_path = value

    @property
    def max_unix_socket_connections(self) -> int:
        return self._max_unix_socket_connections

    @max_unix_socket_connections.setter
    def max_unix_socket_connections(self, value: int) -> None:
        if not isinstance(value, int) or value < 1:
            raise ElhazValidationError(
                f"Invalid max Unix socket connections: '{value}'"
            )
        self._max_unix_socket_connections = value

    @property
    def daemon_logging_path(self) -> Path:
        return self._daemon_logging_path

    @daemon_logging_path.setter
    def daemon_logging_path(self, value: Path) -> None:
        if not value or not isinstance(value, Path):
            raise ElhazValidationError(
                f"Invalid daemon logging path: '{value}'"
            )
        self._daemon_logging_path = value

    @property
    def max_daemon_cache_size(self) -> int:
        return self._max_daemon_cache_size

    @max_daemon_cache_size.setter
    def max_daemon_cache_size(self, value: int) -> None:
        if not isinstance(value, int) or value < 1:
            raise ElhazValidationError(
                f"Invalid max daemon cache size: '{value}'"
            )
        self._max_daemon_cache_size = value


state = Constants()



import json
import os
import signal
import socket
import threading
import time
import uuid
from pathlib import Path

import pytest

from elhaz.constants import Constants
from elhaz.daemon import Client, DaemonService, Server
from elhaz.exceptions import ElhazDaemonError
from elhaz.models import ResponseModel


class FakeSTSClient:
    def get_caller_identity(self) -> dict:
        return {
            "Account": "123456789012",
            "Arn": ("arn:aws:sts::123456789012:assumed-role/demo/session"),
            "UserId": "AROATEST:session",
        }


class FakeRefreshableSession:
    credentials = {
        "access_key": "AKIATEST",
        "secret_key": "secret",
        "token": "token",
        "expiry_time": "2030-01-01T00:00:00Z",
    }

    def client(self, service_name: str) -> FakeSTSClient:
        return FakeSTSClient()


class FakeSession:
    def __init__(self, name: str) -> None:
        self.name = name
        self.session = FakeRefreshableSession()


def _socket_path(prefix: str = "elhaz-test") -> Path:
    return Path("/tmp") / f"{prefix}-{uuid.uuid4().hex}.sock"


def _wait_for_socket(path: Path, timeout: float = 2.0) -> None:
    deadline = time.time() + timeout
    while time.time() < deadline:
        if path.exists():
            return
        time.sleep(0.01)
    raise AssertionError(f"Socket {path!r} did not appear in time.")


@pytest.fixture()
def constants() -> Constants:
    c = Constants()
    c.socket_path = _socket_path()
    return c


@pytest.fixture()
def service(monkeypatch) -> DaemonService:
    monkeypatch.setattr("elhaz.daemon.Session", FakeSession)
    return DaemonService()


@pytest.fixture()
def running_server(constants, service):
    server = Server(constants, service)
    thread = threading.Thread(target=server.run, daemon=True)
    thread.start()
    _wait_for_socket(constants.socket_path)
    yield server, constants
    server.stop()
    thread.join(timeout=2)


def test_add_session(running_server):
    server, constants = running_server
    with Client(constants) as client:
        response = client.send("add", {"config": "demo"})
    assert response.ok is True
    assert response.data == {"config": "demo"}
    assert response.error is None


def test_list_sessions(running_server):
    server, constants = running_server
    with Client(constants) as client:
        client.send("add", {"config": "demo"})
    with Client(constants) as client:
        response = client.send("list")
    assert response.ok is True
    assert response.data == ["demo"]
    assert response.version == 1


def test_credentials(running_server):
    server, constants = running_server
    with Client(constants) as client:
        client.send("add", {"config": "demo"})
    with Client(constants) as client:
        response = client.send("credentials", {"config": "demo"})
    assert response.ok is True
    assert response.data["access_key"] == "AKIATEST"
    assert response.data["secret_key"] == "secret"


def test_whoami(running_server):
    server, constants = running_server
    with Client(constants) as client:
        client.send("add", {"config": "demo"})
    with Client(constants) as client:
        response = client.send("whoami", {"config": "demo"})
    assert response.ok is True
    assert response.data == {
        "Account": "123456789012",
        "Arn": "arn:aws:sts::123456789012:assumed-role/demo/session",
        "UserId": "AROATEST:session",
    }


def test_remove_session(running_server):
    server, constants = running_server
    with Client(constants) as client:
        client.send("add", {"config": "demo"})
    with Client(constants) as client:
        response = client.send("remove", {"config": "demo"})
    assert response.ok is True
    assert response.data == {"config": "demo"}
    with Client(constants) as client:
        list_response = client.send("list")
    assert list_response.data == []


def test_request_id_is_echoed(running_server):
    """The response request_id must match the one sent in the request."""
    server, constants = running_server
    req_id = str(uuid.uuid4())

    raw_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    raw_sock.connect(str(constants.socket_path))
    conn_file = raw_sock.makefile("rwb")

    request = {
        "version": 1,
        "request_id": req_id,
        "action": "list",
        "payload": {},
    }
    conn_file.write(json.dumps(request).encode("utf-8") + b"\n")
    conn_file.flush()

    raw = conn_file.readline()
    conn_file.close()
    raw_sock.close()

    response = ResponseModel.model_validate_json(raw.rstrip())
    assert str(response.request_id) == req_id


def test_invalid_json_returns_structured_400(running_server):
    server, constants = running_server

    raw_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    raw_sock.connect(str(constants.socket_path))
    conn_file = raw_sock.makefile("rwb")
    conn_file.write(b"not-json\n")
    conn_file.flush()

    raw = conn_file.readline()
    conn_file.close()
    raw_sock.close()

    response = ResponseModel.model_validate_json(raw.rstrip())
    assert response.ok is False
    assert response.error is not None
    assert response.error.code == 400


def test_invalid_action_returns_400(running_server):
    """An action not in the Literal fails Pydantic validation -> 400."""
    server, constants = running_server

    raw_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    raw_sock.connect(str(constants.socket_path))
    conn_file = raw_sock.makefile("rwb")

    request = {
        "version": 1,
        "request_id": str(uuid.uuid4()),
        "action": "explode",
        "payload": {},
    }
    conn_file.write(json.dumps(request).encode("utf-8") + b"\n")
    conn_file.flush()

    raw = conn_file.readline()
    conn_file.close()
    raw_sock.close()

    response = ResponseModel.model_validate_json(raw.rstrip())
    assert response.ok is False
    assert response.error.code == 400  # type: ignore[attr-defined]


def test_missing_payload_config_returns_400(running_server):
    server, constants = running_server
    with Client(constants) as client:
        response = client.send("add", {})
    assert response.ok is False
    assert response.error.code == 400  # type: ignore[attr-defined]


def test_not_found_returns_404(running_server):
    server, constants = running_server
    with Client(constants) as client:
        response = client.send("credentials", {"config": "ghost"})
    assert response.ok is False
    assert response.error.code == 404  # type: ignore[attr-defined]


def test_remove_not_found_returns_404(running_server):
    server, constants = running_server
    with Client(constants) as client:
        response = client.send("remove", {"config": "ghost"})
    assert response.ok is False
    assert response.error.code == 404  # type: ignore[attr-defined]


def test_server_continues_after_error(running_server):
    """A failed request must not affect subsequent clients."""
    server, constants = running_server
    with Client(constants) as client:
        error_response = client.send("credentials", {"config": "ghost"})
    with Client(constants) as client:
        ok_response = client.send("list")
    assert error_response.ok is False
    assert ok_response.ok is True


def test_idle_client_does_not_block_second_client(running_server):
    """An open but idle connection must not block a second client."""
    server, constants = running_server

    idle = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    idle.connect(str(constants.socket_path))

    result: dict[str, ResponseModel] = {}

    def _send_list() -> None:
        with Client(constants) as client:
            result["response"] = client.send("list")

    t = threading.Thread(target=_send_list)
    t.start()
    t.join(timeout=2)
    idle.close()

    assert not t.is_alive(), "Second client thread timed out"
    assert result["response"].ok is True
    assert result["response"].data == []


def test_stop_removes_socket_file(constants, service):
    server = Server(constants, service)
    thread = threading.Thread(target=server.run, daemon=True)
    thread.start()
    _wait_for_socket(constants.socket_path)

    server.stop()
    thread.join(timeout=2)

    assert not constants.socket_path.exists()
    assert not thread.is_alive()


def test_stop_is_idempotent(constants, service):
    server = Server(constants, service)
    thread = threading.Thread(target=server.run, daemon=True)
    thread.start()
    _wait_for_socket(constants.socket_path)

    server.stop()
    server.stop()  # must not raise
    thread.join(timeout=2)

    assert not thread.is_alive()


def test_kill_stops_server(running_server):
    """kill action must shut the server down cleanly."""
    server, constants = running_server

    with Client(constants) as client:
        response = client.send("kill")

    assert response.ok is True

    # Socket file is removed once stop() completes.
    deadline = time.time() + 2.0
    while time.time() < deadline:
        if not constants.socket_path.exists():
            break
        time.sleep(0.01)
    assert not constants.socket_path.exists()


def test_startup_rejects_live_daemon(constants, service):
    """Server.__init__ must raise if a daemon is already listening."""
    server = Server(constants, service)
    thread = threading.Thread(target=server.run, daemon=True)
    thread.start()
    _wait_for_socket(constants.socket_path)

    try:
        with pytest.raises(ElhazDaemonError, match="already running"):
            Server(constants, DaemonService())
    finally:
        server.stop()
        thread.join(timeout=2)


def test_startup_clears_stale_socket(constants, service):
    """Server.__init__ must remove a dead socket file and start cleanly."""
    # Bind to the path and close without unlinking — leaves a stale file.
    stale = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    stale.bind(str(constants.socket_path))
    stale.close()

    assert constants.socket_path.exists()

    server = Server(constants, service)
    thread = threading.Thread(target=server.run, daemon=True)
    thread.start()
    _wait_for_socket(constants.socket_path)
    server.stop()
    thread.join(timeout=2)

    assert not thread.is_alive()


def test_startup_rejects_non_socket_path(constants, service):
    """Server.__init__ must raise if the path is a regular file."""
    constants.socket_path.write_bytes(b"")

    try:
        with pytest.raises(ElhazDaemonError, match="non-socket"):
            Server(constants, service)
    finally:
        constants.socket_path.unlink(missing_ok=True)


def test_client_wraps_connection_error(constants):
    """Client.__init__ must raise ElhazDaemonError when no daemon listens."""
    with pytest.raises(ElhazDaemonError, match="Could not connect"):
        Client(constants)


def test_stop_before_run_does_not_raise(constants, service):
    """stop() must be safe to call before run() is ever invoked.

    Exercises the self._sock None-guard and atexit-registered stop().
    """
    server = Server(constants, service)
    server.stop()  # must not raise
    assert not constants.socket_path.exists()


def test_signal_handlers_unchanged_from_non_main_thread(constants, service):
    """Signal handlers must not be modified when run() runs in a thread.

    All other tests call run() from a daemon thread, so this verifies
    the threading.main_thread() guard is in place.
    """
    before_sigterm = signal.getsignal(signal.SIGTERM)
    before_sigint = signal.getsignal(signal.SIGINT)

    server = Server(constants, service)
    thread = threading.Thread(target=server.run, daemon=True)
    thread.start()
    _wait_for_socket(constants.socket_path)
    server.stop()
    thread.join(timeout=2)

    assert signal.getsignal(signal.SIGTERM) is before_sigterm
    assert signal.getsignal(signal.SIGINT) is before_sigint


def test_sigterm_shuts_down_server(constants, service):
    """SIGTERM triggers clean shutdown when run() is called from main thread.

    Delivers SIGTERM to the current process from a timer thread. The
    signal is received in the main thread (where run() is blocked) and
    the registered handler calls stop(), causing the accept loop to exit.
    Original signal handlers are restored before this test returns.
    """
    server = Server(constants, service)

    # Fire SIGTERM after the server is listening. The 0.5 s accept()
    # timeout means run() exits within ~0.7 s of the signal.
    timer = threading.Timer(0.2, lambda: os.kill(os.getpid(), signal.SIGTERM))
    timer.start()
    try:
        server.run()  # blocks in the main thread
    finally:
        timer.cancel()

    assert not constants.socket_path.exists()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for elhaz.models."""

from __future__ import annotations

from uuid import uuid4

import pytest
from pydantic import ValidationError

from elhaz.models import (
    AssumeRoleModel,
    ConfigModel,
    CredentialProcessModel,
    ErrorModel,
    MFAModel,
    ProvidedContextModel,
    RequestModel,
    ResponseModel,
    SessionModel,
    STSModel,
    TagModel,
)

VALID_ROLE_ARN = "arn:aws:iam::123456789012:role/TestRole"


@pytest.mark.parametrize(
    "action",
    ["add", "credentials", "kill", "list", "remove", "whoami"],
)
def test_request_model_all_valid_actions(action: str) -> None:
    m = RequestModel(request_id=uuid4(), action=action)  # type: ignore[call-arg]
    assert m.action == action


def test_request_model_rejects_unknown_action() -> None:
    with pytest.raises(ValidationError):
        RequestModel(request_id=uuid4(), action="explode")  # type: ignore[call-arg]


def test_request_model_rejects_extra_fields() -> None:
    with pytest.raises(ValidationError):
        RequestModel(request_id=uuid4(), action="list", unexpected_field="x")  # type: ignore[call-arg]


def test_request_model_default_version() -> None:
    m = RequestModel(request_id=uuid4(), action="list")
    assert m.version == 1


def test_request_model_default_payload_is_empty_dict() -> None:
    m = RequestModel(request_id=uuid4(), action="list")
    assert m.payload == {}


def test_request_model_requires_request_id() -> None:
    with pytest.raises(ValidationError):
        RequestModel(action="list")  # type: ignore[call-arg]


def test_request_model_custom_payload() -> None:
    m = RequestModel(
        request_id=uuid4(), action="add", payload={"config": "demo"}
    )
    assert m.payload == {"config": "demo"}


def test_response_model_ok_true_with_data() -> None:
    rid = uuid4()
    m = ResponseModel(request_id=rid, ok=True, data={"foo": "bar"})
    assert m.ok is True
    assert m.data == {"foo": "bar"}
    assert m.error is None


def test_response_model_ok_false_with_error() -> None:
    rid = uuid4()
    m = ResponseModel(
        request_id=rid,
        ok=False,
        error=ErrorModel(code=404, message="not found"),
    )
    assert m.ok is False
    assert m.error is not None
    assert m.error.code == 404
    assert m.data is None


def test_response_model_rejects_extra_fields() -> None:
    with pytest.raises(ValidationError):
        ResponseModel(request_id=uuid4(), ok=True, surprise="oops")  # type: ignore[call-arg]


def test_response_model_default_version() -> None:
    m = ResponseModel(request_id=uuid4(), ok=True)
    assert m.version == 1


def test_error_model_valid() -> None:
    m = ErrorModel(code=400, message="bad request")
    assert m.code == 400
    assert m.message == "bad request"


def test_error_model_extra_forbidden() -> None:
    with pytest.raises(ValidationError):
        ErrorModel(code=500, message="oops", extra_field="x")  # type: ignore[call-arg]


def test_config_model_requires_assume_role() -> None:
    with pytest.raises(ValidationError):
        ConfigModel()  # type: ignore[call-arg]


def test_config_model_minimal() -> None:
    m = ConfigModel(AssumeRole=AssumeRoleModel(RoleArn=VALID_ROLE_ARN))
    assert m.AssumeRole.RoleArn == VALID_ROLE_ARN
    assert m.STS is None
    assert m.MFA is None
    assert m.Session is None


def test_config_model_with_all_optional_sections() -> None:
    m = ConfigModel(
        AssumeRole=AssumeRoleModel(RoleArn=VALID_ROLE_ARN),
        STS=STSModel(region_name="us-east-1"),
        MFA=MFAModel(command="oathtool"),
        Session=SessionModel(region_name="us-west-2"),
    )
    assert m.STS is not None
    assert m.MFA is not None
    assert m.Session is not None


def test_config_model_dump_exclude_none_drops_none_fields() -> None:
    m = ConfigModel(AssumeRole=AssumeRoleModel(RoleArn=VALID_ROLE_ARN))
    dumped = m.model_dump(exclude_none=True)
    assert "AssumeRole" in dumped
    assert "STS" not in dumped
    assert "MFA" not in dumped
    assert "Session" not in dumped


def test_assume_role_model_requires_role_arn() -> None:
    with pytest.raises(ValidationError):
        AssumeRoleModel()  # type: ignore[call-arg]


def test_assume_role_model_with_tags() -> None:
    m = AssumeRoleModel(
        RoleArn=VALID_ROLE_ARN,
        Tags=[TagModel(Key="env", Value="prod")],
    )
    assert m.Tags is not None
    assert len(m.Tags) == 1
    assert m.Tags[0].Key == "env"
    assert m.Tags[0].Value == "prod"


def test_assume_role_model_extra_fields_forbidden() -> None:
    with pytest.raises(ValidationError):
        AssumeRoleModel(RoleArn=VALID_ROLE_ARN, Unknown="x")  # type: ignore[call-arg]


def test_assume_role_model_all_optional_none_by_default() -> None:
    m = AssumeRoleModel(RoleArn=VALID_ROLE_ARN)
    dumped = m.model_dump(exclude_none=True)
    assert list(dumped.keys()) == ["RoleArn"]


def test_assume_role_model_with_provided_contexts() -> None:
    m = AssumeRoleModel(
        RoleArn=VALID_ROLE_ARN,
        ProvidedContexts=[
            ProvidedContextModel(
                ProviderArn="arn:aws:iam::123456789012:oidc-provider/foo",
                ContextAssertion="ctx",
            )
        ],
    )
    assert m.ProvidedContexts is not None
    assert len(m.ProvidedContexts) == 1


def test_credential_process_model_default_version() -> None:
    m = CredentialProcessModel(
        AccessKeyId="AKIA",
        SecretAccessKey="secret",
        SessionToken="token",
        Expiration="2030-01-01T00:00:00Z",
    )
    assert m.Version == 1


def test_credential_process_model_all_fields() -> None:
    m = CredentialProcessModel(
        Version=1,
        AccessKeyId="AKID",
        SecretAccessKey="SAK",
        SessionToken="TOK",
        Expiration="2030-01-01T00:00:00Z",
    )
    assert m.AccessKeyId == "AKID"
    assert m.SecretAccessKey == "SAK"
    assert m.SessionToken == "TOK"
    assert m.Expiration == "2030-01-01T00:00:00Z"


def test_sts_model_all_optional() -> None:
    m = STSModel()
    assert m.region_name is None
    assert m.aws_access_key_id is None


def test_sts_model_with_region() -> None:
    m = STSModel(region_name="eu-west-1")
    assert m.region_name == "eu-west-1"


def test_sts_model_verify_can_be_bool_or_str() -> None:
    assert STSModel(verify=True).verify is True
    assert STSModel(verify="/path/to/ca.pem").verify == "/path/to/ca.pem"


def test_mfa_model_requires_command() -> None:
    with pytest.raises(ValidationError):
        MFAModel()  # type: ignore[call-arg]


def test_mfa_model_default_timeout() -> None:
    m = MFAModel(command="oathtool")
    assert m.timeout == 30


def test_mfa_model_command_can_be_list() -> None:
    m = MFAModel(command=["oathtool", "--totp", "SECRET"])
    assert isinstance(m.command, list)


def test_session_model_all_optional() -> None:
    m = SessionModel()
    dumped = m.model_dump(exclude_none=True)
    assert dumped == {}


def test_session_model_with_region() -> None:
    m = SessionModel(region_name="ap-southeast-1")
    assert m.region_name == "ap-southeast-1"


def test_tag_model_valid() -> None:
    m = TagModel(Key="CostCenter", Value="9999")
    assert m.Key == "CostCenter"
    assert m.Value == "9999"


def test_tag_model_requires_both_fields() -> None:
    with pytest.raises(ValidationError):
        TagModel(Key="only-key")  # type: ignore[call-arg]


def test_provided_context_model_valid() -> None:
    m = ProvidedContextModel(
        ProviderArn="arn:aws:iam::123:oidc-provider/foo",
        ContextAssertion="assertion-value",
    )
    assert m.ProviderArn.startswith("arn:")
    assert m.ContextAssertion == "assertion-value"


def test_provided_context_model_requires_both_fields() -> None:
    with pytest.raises(ValidationError):
        ProvidedContextModel(ProviderArn="arn:x")  # type: ignore[call-arg]
